@@ -10,6 +10,11 @@ use Data::Dumper;
 use POSIX qw//;
 use Encode qw//;
 use XML::LibXML;
+use Interchange::Search::Solr::Response;
+use Lingua::StopWords;
+use Types::Standard qw/ArrayRef HashRef Int Bool/;
+use namespace::clean;
+use HTTP::Response;
 
 =head1 NAME
 
@@ -17,11 +22,11 @@ Interchange::Search::Solr -- Solr query encapsulation
 
 =head1 VERSION
 
-Version 0.09
+Version 0.10
 
 =cut
 
-our $VERSION = '0.09';
+our $VERSION = '0.10';
 
 =head1 DESCRIPTION
 
@@ -182,6 +187,52 @@ wildcard is prepended as well, querying for *1234* instead).
 has wild_matching => (is => 'ro',
                       default => sub { 0 });
 
+=head2 stop_words_langs
+
+The languages for which we should build the stop word list. It
+defaults to:
+
+ [ 'en' ]
+
+New in 0.10. To revert to the old behaviour (no filtering of
+stopwords), pass an empty arrayref.
+
+=cut
+
+has stop_words => (is => 'lazy', isa => HashRef);
+
+has stop_words_langs => (is => 'ro', default => sub { [qw/en/ ] }, isa => ArrayRef);
+
+sub _build_stop_words {
+    my $self = shift;
+    my @stopwords;
+    foreach my $lang (@{ $self->stop_words_langs }) {
+        if (my $stops = Lingua::StopWords::getStopWords($lang, 'UTF-8')) {
+            push @stopwords, keys %$stops;
+        }
+    }
+    my %out = map { $_ => 1 } @stopwords;
+    return \%out;
+}
+
+=head2 min_chars
+
+Minimum characters for filtering the search terms. Default to 3.
+
+New in 0.10. To revert to the old behaviour, set it to 0.
+
+=head2 permit_empty_search
+
+By default, empty searches are not executed. You can permit them
+setting this accessor to 1. The module will reset it to 0 when the
+search is executed.
+
+=cut
+
+has min_chars => (is => 'ro', isa => Int, default => sub { 3 });
+
+has permit_empty_search => (is => 'rw', isa => Bool, default => sub { 0 });
+
 has search_fields => (is => 'ro',
                       default => sub {
                           return [
@@ -211,7 +262,7 @@ has page => (is => 'rw',
              default => sub { 1 });
 
 has filters => (is => 'rw',
-                isa => sub { die "not an hashref" unless ref($_[0]) eq 'HASH' },
+                isa => HashRef,
                 default => sub { return {} },
                );
 
@@ -332,7 +383,7 @@ sub search {
     }
     my @terms;
     if ($string) {
-        @terms = grep { $_ } split(/\s+/, $string);
+        @terms = grep { $self->_term_is_good($_) } split(/\s+/, $string);
     }
     $self->search_terms(\@terms);
     $self->search_structure($structure);
@@ -342,7 +393,7 @@ sub search {
 sub _do_search {
     my $self = shift;
 
-    my @terms = grep { /\w/ } @{ $self->search_terms };
+    my @terms = grep { $self->_term_is_good($_) } @{ $self->search_terms };
 
     my $query = '';
     my $wild_match = '';
@@ -375,6 +426,19 @@ If no query is provided, a wildcard search is performed.
 
 =cut
 
+sub _search_is_empty {
+    my $self = shift;
+    my @terms = @{ $self->search_terms };
+    my %filters = %{ $self->filters };
+    my $structure = $self->search_structure;
+    if (@terms || %filters || $structure) {
+        return 0;
+    }
+    else {
+        return 1;
+    }
+}
+
 sub execute_query {
     my ($self, $query) = @_;
     my $querystring = '*';
@@ -402,10 +466,22 @@ sub execute_query {
     }
     # save the debug info
     $self->_set_search_string($querystring);
-    my %params = $self->construct_params;
-    my $res = $self->solr_object->search($querystring, \%params);
-    $self->_set_response($res);
-    return $res;
+
+    my $our_res;
+    unless ($self->permit_empty_search) {
+        if ($self->_search_is_empty) {
+            $our_res = Interchange::Search::Solr::Response->new(HTTP::Response->new(404));
+            $our_res->error('empty_search');
+        }
+    }
+    unless ($our_res) {
+        my %params = $self->construct_params;
+        my $res = $self->solr_object->search($querystring, \%params);
+        $our_res = Interchange::Search::Solr::Response->new($res->raw_response);
+    }
+    $self->_set_response($our_res);
+    $self->permit_empty_search(0);
+    return $our_res;
 }
 
 =head2 construct_params
@@ -659,7 +735,7 @@ sub add_terms_to_url {
     die "Bad usage" unless defined $url;
     $self->_parse_url($url);
     return $url unless @other_terms;
-    my @additional_terms = grep { $_ } @other_terms;
+    my @additional_terms = grep { $self->_term_is_good($_) } @other_terms;
     my @terms = @{ $self->search_terms };
     push @terms, @additional_terms;
     $self->search_terms(\@terms);
@@ -732,7 +808,8 @@ sub _parse_url {
             push @terms, $chunk;
         }
     }
-    $self->search_terms(\@terms);
+    # filter the terms
+    $self->search_terms([ grep { $self->_term_is_good($_) } @terms ]);
     $self->filters(\%filters);
 }
 
@@ -857,27 +934,37 @@ The structure looks like this:
 
  {
    first => 'words/bla' || undef,
+   first_page => 1 || undef,
    last  => 'words/bla/page/5' || undef,
-   next => 'words/bla/page/3' || undef,
-   previous => 'words/bla/page/5' || undef,
+   last_page => 5 || undef,
+   next => 'words/bla/page/5' || undef,
+   next_page => 5 || undef
+   previous => 'words/bla/page/3' || undef,
+   previous_page => 3 || undef,
    pages => [
              {
+              name => 1,
               url => 'words/bla/page/1',
              },
              {
+              name => 2,
               url => 'words/bla/page/2',
              },
              {
+              name => 3,
               url => 'words/bla/page/3',
              },
              {
+              name => 4,
               url => 'words/bla/page/4',
               current => 1,
              },
              {
+              name => 5,
               url => 'words/bla/page/5',
              },
             ]
+   total_pages => 5
  }
 
 =cut
@@ -912,9 +999,11 @@ sub paginator {
         }
         elsif ($position == 1) {
             $pager{next} = $url;
+            $pager{next_page} = $count;
         }
         elsif ($position == -1) {
             $pager{previous} = $url;
+            $pager{previous_page} = $count;
         }
         push @{$pager{items}}, $item;
     }
@@ -922,11 +1011,14 @@ sub paginator {
         $pager{last} = $self->url_builder($self->search_terms,
                                           $self->filters,
                                           $total_pages);
+        $pager{last_page} = $total_pages;
     }
     if ($page != 1) {
         $pager{first} = $self->url_builder($self->search_terms,
                                            $self->filters, 1);
+        $pager{first_page} = 1;
     }
+    $pager{total_pages} = $total_pages;
     return \%pager;
 }
 
@@ -1061,6 +1153,26 @@ sub remove_word_links {
     return @out;
 }
 
+sub _term_is_good {
+    my ($self, $term) = @_;
+    if ($term && $term =~ /\w/) {
+        if ($self->stop_words->{lc($term)}) {
+            return 0;
+        }
+        if ($self->min_chars > 1) {
+            my $re = "\\w.*" x $self->min_chars;
+            if ($term =~ m/$re/) {
+                return 1;
+            }
+            else {
+                return 0;
+            }
+        }
+        return 1;
+    }
+    return 0;
+}
+
 
 =head1 AUTHOR
 
